/*==============================================================================*/
/*  Copyright (C) 2009-2016, Heartland Data inc. All Rights Reserved.           */
/*                                                                              */
/*  Title  :   GPIO 4bit Driver                                                 */
/*  EventID:   Insert Kernelinfo Test Point                                     */
/*  AppVer :   DT10 `Ver10.51                                                  */
/*  FileID :   33b                                                              */
/*  Version:   3.2                                                              */
/*  Author :   HLDC                                                             */
/*==============================================================================*/

/*==============================================================================*/
/*  Please customize the code for your environment.                             */
/*==============================================================================*/

/*==============================================================================*/
/*  Desc:   Header for PORT Control                                             */
/*==============================================================================*/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <linux/version.h>

#define LICENSE "GPL v2"

/*==============================================================================*/
/*  Macro:  DT_UINT                                                             */
/*  Desc:   Please change Test Point argument type for DT10 Project setting.    */
/*==============================================================================*/
#define	DT_UINT unsigned int

/*==============================================================================*/
/*  Macro:  PROC_NAME                                                           */
/*  Desc:   Register Proc File name for Linux.                                  */
/*==============================================================================*/
#define	PROC_NAME "dt10proc"

/*==============================================================================*/
/*  Struct:  DT10_DATA                                                          */
/*  Desc:    Test Point Data Struct                                             */
/*==============================================================================*/
struct DT10_DATA {
	char tp_func;
	DT_UINT addr;
	DT_UINT dat;
	DT_UINT size;
	unsigned char value[256];
};

/*==============================================================================*/
/*  Desc:    DT10_DATA.tp_func ID                                               */
/*==============================================================================*/
#define TP_BusOut		'2'
#define TP_MemoryOutput		'3'
#define TP_BusOutFast		'4'
#define TP_MemoryOutputFast		'5'
#define TP_EventTrigger		'6'
#define TP_EventTriggerFast		'7'
#define TP_EventTrigger32		'8'
#define TP_EventTrigger32Fast		'9'

/*==============================================================================*/
/*  Desc:   Test Point Parametaer                                               */
/*==============================================================================*/
#define		DT_VARIABLE_BIT			0x02
#define		DT_EVTTRG_BIT			0x08
#define		DT_VARIABLE_FAST_BIT		0x01
#define		DT_EVTTRG_FAST_BIT		0x02

/*==============================================================================*/
/*  Macro:  DT_ADD_CPU_INFO                                                     */
/*  Desc:   Please change Test Point argument type for DT10 Project setting.    */
/*==============================================================================*/
#define DT_ADD_CPU_INFO		0

/*==============================================================================*/
/*  Macro:  DT_USE_SPINLOCK                                                     */
/*  Desc:   Please set 1, when use spin_lock_irqsave function.                  */
/*==============================================================================*/
#include <linux/spinlock.h>
static DEFINE_SPINLOCK(tplock);
unsigned long flags = 0;
#define DT10_LOCK	spin_lock_irqsave(&tplock, flags)
#define DT10_UNLOCK	spin_unlock_irqrestore(&tplock, flags)

/*==============================================================================*/
/*  Desc:   GPIO Control setting                                                */
/*==============================================================================*/
#define BCM2708_PERI_BASE        0x20000000
#define GPIO_BASE                (BCM2708_PERI_BASE + 0x200000)

#define	CS_BIT		7	/* CS */
#define	CLK_BIT		25	/* CLK */
#define	DAT0_BIT	8	/* DAT0 */
#define	DAT1_BIT	9	/* DAT1 */
#define	DAT2_BIT	10	/* DAT2 */
#define	DAT3_BIT	11	/* DAT3 */

void *iop = NULL;

#define	INP_GPIO(g) iowrite32(\
	ioread32( (void*)((int *)iop+(g)/10) ) & ~(7<<(((g)%10)*3)), \
		  (void*)((int *)iop+(g)/10) )
#define OUT_GPIO(g) iowrite32(\
	ioread32( (void*)((int *)iop+(g)/10) ) | (1<<(((g)%10)*3)), \
		  (void*)((int *)iop+(g)/10) )
#define	GPIO_SET(g) iowrite32( 1<<(g), (void*)((int *)iop+7) )
#define	GPIO_CLR(g) iowrite32( 1<<(g), (void*)((int *)iop+10) )
#define	GPIO_GET ioread32((void*)((int *)iop+13))

#define	CLK_CS_HI	{			\
	GPIO_SET(CS_BIT);			\
	GPIO_SET(CLK_BIT);			\
}

#define	CS_HI	{				\
	GPIO_SET(CS_BIT);			\
}
#define	CS_LOW	{				\
	GPIO_CLR(CS_BIT);			\
}

#define	CLK_HI	{				\
	GPIO_SET(CLK_BIT);			\
}

#define	CLK_LOW	{				\
	GPIO_CLR(CLK_BIT);			\
}

#define DATA_OUT(x) {			\
	( (x) & 0x1) ? GPIO_SET(DAT0_BIT) : GPIO_CLR(DAT0_BIT);	\
	( (x) & 0x2) ? GPIO_SET(DAT1_BIT) : GPIO_CLR(DAT1_BIT);	\
	( (x) & 0x4) ? GPIO_SET(DAT2_BIT) : GPIO_CLR(DAT2_BIT);	\
	( (x) & 0x8) ? GPIO_SET(DAT3_BIT) : GPIO_CLR(DAT3_BIT);	\
}

static int init = 0;	/* init flag */
static unsigned int start_trace = 0;
/*==============================================================================*/
/*  Func:   _TP_BusPortInit                                                     */
/*  Desc:   Please describe the code to initializes ports.                      */
/*==============================================================================*/
static void _TP_BusPortInit(void)
{
	unsigned int reg;

	if (init == 0) {
		iop = ioremap(GPIO_BASE,16);
		if (iop == NULL) {
			printk("dt10drvspi: init err\n");
			return;
		}
		OUT_GPIO(CS_BIT);
		OUT_GPIO(CLK_BIT);
		OUT_GPIO(DAT0_BIT);
		OUT_GPIO(DAT1_BIT);
		OUT_GPIO(DAT2_BIT);
		OUT_GPIO(DAT3_BIT);

		CLK_CS_HI;

		init = 1;
		return;
	}
}

/*==============================================================================*/
/*  Func:   _TP_BusOutDrv                                                       */
/*  Desc:   Test Point Output Function.                                         */
/*==============================================================================*/
static inline void _TP_BusOutDrv( DT_UINT addr, DT_UINT data )
{
#if DT_ADD_CPU_INFO
	int cpu_id = 0;
	cpu_id = smp_processor_id();
	if ((cpu_id < 0) || (cpu_id > 3)) {
		cpu_id = 0;
	}
	addr &= ~(0x3 << 18); 
	addr |= cpu_id << 18;
#endif
	CS_LOW;

	DATA_OUT(data >> 12);		/* bit15-12 */
	CLK_LOW;
	DATA_OUT(data >> 8);		/* bit11-8 */
	CLK_HI;
	DATA_OUT(data >> 4);		/* bit7-4 */
	CLK_LOW;
	DATA_OUT(data);				/* bit3-0 */
	CLK_HI;

	if( addr >= 0x00010000 )
	{
		DATA_OUT(addr >> 20);		/* bit23-20 */
		CLK_LOW;
		DATA_OUT(addr >> 16);		/* bit19-16 */
		CLK_HI;
	}
	if( addr >= 0x00000100 )
	{
		DATA_OUT(addr >> 12);		/* bit15-12 */
		CLK_LOW;
		DATA_OUT(addr >> 8);		/* bit11-8 */
		CLK_HI;
	}
	DATA_OUT(addr >> 4);			/* bit7-4 */
	CLK_LOW;
	DATA_OUT(addr);					/* bit3-0 */
	CLK_HI;

	CS_HI;
}

/*==============================================================================*/
/*  Func:   _TP_MemoryOutDrv                                                    */
/*  Desc:   Value Output Function.                                              */
/*==============================================================================*/
static inline void _TP_MemoryOutDrv(unsigned char data)
{
	DATA_OUT((data >> 4) & 0xFF);			/* bit7-4 */
	CLK_LOW;
	DATA_OUT((data) & 0xFF);			/* bit3-0 */
	CLK_HI;
}

/*==============================================================================*/
/*  Func:   _TP_BusOutFastDrv                                                   */
/*  Desc:   Test Point Output Function.                                         */
/*==============================================================================*/
static void _TP_BusOutFastDrv(DT_UINT data, DT_UINT bit_num)
{
	CS_LOW;
	switch (bit_num) {
	case 12:
	case 11:
	case 10:
	case 9:
		DATA_OUT((data >> 8) & 0xFF);
		CLK_LOW;
	case 8:
	case 7:
	case 6:
	case 5:
		DATA_OUT((data >> 4) & 0xFF);
		CLK_LOW;
	default:
		DATA_OUT((data) & 0xFF);
		CLK_LOW;
		break;

	}
	CS_HI;
	CLK_HI;
}

/**********************************kernel TestPoint API**********************************/

/*==============================================================================*/
/*  Func:   _TP_BusOut                                                          */
/*  Desc:   Called by Test Point                                                */
/*==============================================================================*/
void _TP_BusOut( DT_UINT addr, DT_UINT dat )
{
	if (start_trace != 1) {
		return;
	}
	DT10_LOCK;
	_TP_BusOutDrv(addr, dat);
	DT10_UNLOCK;
}
EXPORT_SYMBOL(_TP_BusOut);

/*==============================================================================*/
/*  Func:   _TP_MemoryOutput                                                    */
/*  Desc:   Called by Variable TestPoint                                        */
/*==============================================================================*/
void _TP_MemoryOutput( DT_UINT addr, DT_UINT dat, void *value, DT_UINT size )
{
	int	i;
	unsigned char *p = (unsigned char *)value;

	if (start_trace != 1) {
		return;
	}

	if (size > 256) size = 256;
	DT10_LOCK;
	_TP_BusOutDrv((addr | DT_VARIABLE_BIT), dat);
	CS_LOW;
	_TP_MemoryOutDrv(size);
	for (i = 0; i < size; i++) {
		_TP_MemoryOutDrv(p[i]);
	}
	CS_HI;
	DT10_UNLOCK;
}
EXPORT_SYMBOL(_TP_MemoryOutput);

/*==============================================================================*/
/*  Func:   _TP_BusOutFast                                                      */
/*  Desc:   Called by Fast Test Point                                           */
/*==============================================================================*/
void _TP_BusOutFast( DT_UINT step, DT_UINT bit_num )
{
	if (start_trace != 1) {
		return;
	}

	step = step << 2;
	bit_num = bit_num + 2;
	DT10_LOCK;
	_TP_BusOutFastDrv(step, bit_num);
	DT10_UNLOCK;
}
EXPORT_SYMBOL(_TP_BusOutFast);

/*==============================================================================*/
/*  Func:   _TP_MemoryOutputFast                                                */
/*  Desc:   Called by Variable Fast Test Point                                  */
/*==============================================================================*/
void _TP_MemoryOutputFast( DT_UINT step, DT_UINT bit_num, void *value, DT_UINT size)
{
	int	i;
	unsigned char *p = (unsigned char *)value;

	if (start_trace != 1) {
		return;
	}

	step = (step << 2) + DT_VARIABLE_FAST_BIT;
	bit_num = bit_num + 2;
	if (size > 256) size = 256;
	DT10_LOCK;
	_TP_BusOutFastDrv(step, bit_num);
	CS_LOW;
	_TP_MemoryOutDrv(size);
	for (i = 0; i < size; i++) {
		_TP_MemoryOutDrv(p[i]);
	}
	CS_HI;
	DT10_UNLOCK;
}
EXPORT_SYMBOL(_TP_MemoryOutputFast);

/*==============================================================================*/
/*  Func:   _TP_BusKernelInfo                                                   */
/*  Desc:   Called by KernelInfo Test Point.                                    */
/*==============================================================================*/
void _TP_BusKernelInfo(DT_UINT addr, DT_UINT dat, DT_UINT process_id, DT_UINT thread_group_id, unsigned char* command)
{
	int i;
	unsigned char *p = command;

	if (start_trace != 1) {
		return;
	}

	DT10_LOCK;
	_TP_BusOutDrv((addr | DT_VARIABLE_BIT | DT_EVTTRG_BIT), dat);
	CS_LOW;
	_TP_MemoryOutDrv(process_id);
	_TP_MemoryOutDrv((process_id >> 8));
	_TP_MemoryOutDrv((process_id >> 16));
	_TP_MemoryOutDrv((process_id >> 24));
	_TP_MemoryOutDrv(20);
	_TP_MemoryOutDrv(thread_group_id);
	_TP_MemoryOutDrv((thread_group_id >> 8));
	_TP_MemoryOutDrv((thread_group_id >> 16));
	_TP_MemoryOutDrv((thread_group_id >> 24));
	for (i = 0; i < 16; i++) {
		_TP_MemoryOutDrv(p[i]);
	}
	CS_HI;
	DT10_UNLOCK;
}
EXPORT_SYMBOL(_TP_BusKernelInfo);

/********************************************************************************/

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static ssize_t dt10_proc_write(struct file *filp, const char __user *buf, size_t len, loff_t *data)
#else
static int dt10_proc_write(struct file *filp, const char *buf, unsigned long len, void *data)
#endif
{
	struct DT10_DATA dt10_data;

	switch (buf[0]) {
	case '0':
		start_trace = 0;
		printk( KERN_INFO "[DT10] Trace Stop.\n" );
		break;
	case '1':
		start_trace = 1;
		printk( KERN_INFO "[DT10] Trace Start.\n" );
		break;
	case TP_BusOut:
		if (copy_from_user(&dt10_data, buf, len)) {
			return len;
		}
		_TP_BusOut(dt10_data.addr, dt10_data.dat);
		break;
	case TP_MemoryOutput:
		if (copy_from_user(&dt10_data, buf, len)) {
			return len;
		}
		_TP_MemoryOutput(dt10_data.addr, dt10_data.dat, dt10_data.value, dt10_data.size);
		break;
	case TP_BusOutFast:
		if (copy_from_user(&dt10_data, buf, len)) {
			return len;
		}
		_TP_BusOutFast(dt10_data.addr, dt10_data.dat);
		break;
	case TP_MemoryOutputFast:
		if (copy_from_user(&dt10_data, buf, len)) {
			return len;
		}
		_TP_MemoryOutputFast(dt10_data.addr, dt10_data.dat, dt10_data.value, dt10_data.size);
		break;
	default:
		printk( KERN_INFO "[DT10] Trace Command is 0-5.\n" );
		break;
	}

	return len;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static ssize_t dt10_proc_read(struct file *filp, char __user *buf, size_t count, loff_t *data)
#else
static int dt10_proc_read(char *buf, char **start, off_t off, int count, int *eof, void *data)
#endif
{
#if !(LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
	*eof = 1;
#endif
	return count;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static const struct file_operations dt_proc_fops = {
	.write = dt10_proc_write,
	.read = dt10_proc_read,
};
#endif

static int __init dt10_proc_init(void)
{
	struct proc_dir_entry* entry;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
	entry = proc_create(PROC_NAME, 666, NULL, &dt_proc_fops);
	if ( !entry ) {
		printk( KERN_ERR "proc entry fail\n" );
		return -EBUSY;
	}
#else
	entry = create_proc_entry(PROC_NAME, 666, NULL);
	entry->read_proc = dt10_proc_read;
	entry->write_proc = dt10_proc_write;
#endif
	_TP_BusPortInit();
	printk(KERN_INFO "dt10proc load OK\n");
	if(init == 1){
		start_trace = 1;
	}
	return 0;
}

static void __exit dt10_proc_exit(void)
{
	remove_proc_entry(PROC_NAME, NULL);
	if (init != 0) {
		start_trace = 0;
		init = 0;
		iounmap(iop);
	}
	printk(KERN_INFO "dt10proc unload\n");
}

module_init(dt10_proc_init);
module_exit(dt10_proc_exit);

MODULE_DESCRIPTION("dt10proc");
MODULE_LICENSE(LICENSE);

/********************************************************************************/

